package gs.config.v0

import cats.data.NonEmptyList
import cats.effect.Sync
import cats.syntax.all.*
import gs.config.v0.audit.ConfigManifest
import gs.config.v0.audit.ConfigQueryResult
import gs.config.v0.source.ConfigSource
import gs.config.v0.source.EnvironmentConfigSource
import gs.config.v0.source.MemoryConfigSource

/** Implementation of [[gs.config.Configuration]] that tracks every call to
  * `getValue` and reports on whether the query succeeded or failed.
  *
  * @param sources
  *   List of sources to use, in order, when resolving keys.
  * @param manifest
  *   The manifest used to audit configuration loaded by this class.
  */
final class AuditedConfiguration[F[_]: Sync](
  val sources: NonEmptyList[ConfigSource[F]],
  val manifest: ConfigManifest[F]
) extends BaseConfiguration[F]:
  import AuditedConfiguration.Acc

  override def getValue[A: Configurable](
    key: ConfigKey[A]
  ): F[Either[ConfigError, A]] =
    findAndParse(key)
      .map { acc =>
        acc -> (
          acc.result match
            case None         => handleMissingValue(key)
            case Some(result) => parse(key, result, acc.lastAttemptedSource)
        )
      }
      .flatMap { case (acc, result) => audit(key, result, acc).as(result) }

  private def findAndParse[A: Configurable](
    key: ConfigKey[A]
  ): F[Acc] =
    sources.foldLeft(Acc.empty[F]) {
      (
        acc,
        src
      ) =>
        acc.flatMap { data =>
          if data.hasResult then Sync[F].pure(data)
          else
            src.getValue(key).map {
              case None    => data.appendSource(src.name)
              case Some(v) => data.appendSource(src.name).withResult(v)
            }
        }
    }

  private def audit(
    key: ConfigKey[?],
    result: Either[ConfigError, ?],
    acc: Acc
  ): F[Unit] =
    result match
      case Left(error) =>
        manifest.record(
          name = key.name,
          queryResult = ConfigQueryResult.Failure(
            sources = acc.sources.toList,
            error = error
          )
        )
      case Right(_) =>
        manifest.record(
          name = key.name,
          queryResult = ConfigQueryResult.Success(
            source = acc.lastAttemptedSource,
            rawValue = acc.result.getOrElse("")
          )
        )

object AuditedConfiguration:

  /** Instantiate a new [[AuditedConfiguration]] based on the given list of
    * sources.
    *
    * @param sources
    *   The list of [[ConfigSource]] backing this configuration.
    * @return
    *   The new [[Configuration]] instance.
    */
  def forSources[F[_]: Sync](
    sources: NonEmptyList[ConfigSource[F]]
  ): F[Configuration[F]] =
    ConfigManifest
      .standard[F]
      .map(manifest =>
        new AuditedConfiguration[F](
          sources = sources,
          manifest = manifest
        )
      )

  def forSource[F[_]: Sync](source: ConfigSource[F]): Builder[F] =
    Builder[F](NonEmptyList.of(source))

  def forEnvironmentSource[F[_]: Sync]: Builder[F] =
    Builder[F](NonEmptyList.of(new EnvironmentConfigSource[F]))

  def forMemorySource[F[_]: Sync](configs: Map[String, String]): Builder[F] =
    Builder[F](NonEmptyList.of(new MemoryConfigSource[F](configs)))

  case class Builder[F[_]: Sync](sources: NonEmptyList[ConfigSource[F]]):

    def withEnvironmentSource(): Builder[F] =
      copy(sources = this.sources.append(new EnvironmentConfigSource[F]))

    def withMemorySource(configs: Map[String, String]): Builder[F] =
      copy(sources = this.sources.append(new MemoryConfigSource[F](configs)))

    def withSource(source: ConfigSource[F]): Builder[F] =
      copy(sources = this.sources.append(source))

    def build(): F[Configuration[F]] =
      ConfigManifest
        .standard[F]
        .map(manifest =>
          new AuditedConfiguration[F](
            sources = sources,
            manifest = manifest
          )
        )

  /** Internal class for accumulating a result across sources.
    *
    * @param sources
    *   The seen sources.
    * @param result
    *   The result, or `None` if no result has been found.
    */
  private case class Acc(
    sources: Vector[String],
    result: Option[String]
  ):
    def hasResult: Boolean = result.isDefined

    def lastAttemptedSource: String = sources.lastOption.getOrElse("")

    def withResult(r: String): Acc =
      if hasResult then this else this.copy(sources, Some(r))

    def appendSource(s: String): Acc =
      if hasResult then this else this.copy(sources.appended(s), result)

  private object Acc:
    def empty[F[_]: Sync]: F[Acc] = Sync[F].pure(Acc(Vector.empty, None))

end AuditedConfiguration
