package gs.config.v0

/** Uniquely names some piece of configuration. This structure does _not_
  * attempt to support every possible use case, but supports some common cases
  * for users that follow some basic rules. Please review conversion functions
  * to review the rules for each supported conversion.
  *
  * ## Recommended Naming Convention
  *
  * Use either property form (`foo.bar.baz-qwop`) or environment variable form
  * (`FOO_BAR_BAZ_QWOP`). Do not use dashes (`-`) in properties if you need to
  * convert bidirectionally.
  *
  * ## Supported Conversions
  *
  *   - `toRawString`
  *   - `toEnvironmentVariable`
  *   - `toProperty`
  */
opaque type ConfigName = String

object ConfigName:

  /** Instantiate a new `ConfigName`. This function trims all leading and
    * trailing whitespace.
    *
    * @param name
    *   The raw string value of the name.
    * @return
    *   New `ConfigName`.
    */
  def apply(name: String): ConfigName = name.trim()

  given CanEqual[ConfigName, ConfigName] = CanEqual.derived

  extension (name: ConfigName)
    /** Extract the unmodified string that backs this name.
      *
      * @return
      *   The underlying string representation of the name.
      */
    def toRawString(): String = name

    /** Express this name as an environment variable.
      *
      *   - Replaces all alphanumeric characters with `_`.
      *   - Converts all characters to upper case.
      *
      * @return
      *   The environment variable form of this name.
      */
    def toEnvironmentVariable(): String =
      name.replaceAll("[^0-9a-zA-Z_]", "_").toUpperCase()

    /** Express this name as a property.
      *
      *   - Replaces all underscore characters with `.`.
      *   - Converts all characters to lower case.
      *
      * @return
      *   The property form of this name.
      */
    def toProperty(): String =
      name.replaceAll("_", ".").toLowerCase()

end ConfigName
