package gs.config.v0

import cats.effect.Sync
import gs.config.v0.source.ConfigSource

/** Interface for loading configuration. This type should not be used for any
  * sensitive configuration such as secrets or private keys.
  */
trait Configuration[F[_]]:
  /** Retrieve a value based on some key.
    *
    * @param key
    *   The key that identifies the piece of configuration to retrieve.
    * @return
    *   The value, or an error if no value is present.
    */
  def getValue[A: Configurable](key: ConfigKey[A]): F[Either[ConfigError, A]]

object Configuration:

  /** Start building a new [[AuditedConfiguration]].
    *
    * @param source
    *   The first source relevant to the configuration.
    * @return
    *   Fluent builder for the [[AuditedConfiguration]] class.
    */
  def audited[F[_]: Sync](source: ConfigSource[F])
    : AuditedConfiguration.Builder[F] =
    AuditedConfiguration.forSource[F](source)

  /** Create a new [[AuditedConfiguration]] that reads from the environment.
    *
    * @return
    *   The new [[Configuration]].
    */
  def auditedEnvironmentOnly[F[_]: Sync]: F[Configuration[F]] =
    AuditedConfiguration.forEnvironmentSource[F].build()

end Configuration
